import std.stdio;
import std.conv;
import std.string;
import std.file;
import core.thread;
import std.concurrency;
import std.date;

shared ubyte logLevel = 2;

// Change this as needed.
shared pollTimeout = 10_000_000; // 10 Seconds.
const d_time ticksPerHour = makeTime(1, 0, 0, 0);
d_time tzoffset = 0;
shared bool autoflush = false;

shared Tid tid;

/**
    This class allows a developer to add logging to their code and capture
    varying types of output to a user-defined logfile.

    The logger periodically checks the configuration file for updates.
    Currently, only the log level is modifed to allow the user to change the
    log level while the application is running.  This way, the user can change
    the level to capture more information if errors start to occur.

    Log levels are ordered 1 through 5 (Fatal through Trace)

    Logger settings are stored in a file called logger.conf that sits in the
    same directory as the application.  An example follows:
    ---
    level=5
    tzoffset=-5
    ---

    The tzoffset setting is needed as currently std.date currently appears to
    not handle timezones correctly.  Hopefully this will be removed in the
    future.
 */
class Logger
{
    Thread configPoller;
    string location;

    /**
        The constructor for the Logger class.

        This only creates a new instance of the class and sets the location.

        Params:
            loc = The location where the class was instanciated, such as a
            file name or a class name.

        Examples:
        ---
            Logger lgr = new Logger(__FILE__);
        ---
     */
    this(string loc)
    {
        this.location = loc;
    }

    this(string loc, void function() logWriter, bool autoFlush = false)
    {
        version(nologger)
        {
        }
        else
        {
            this.location = loc;
            if (configPoller is null)
            {
                autoflush = autoFlush;
                string[] logFile = splitlines(readText("logger.conf"));
                string[] currConfig;

                foreach (line; logFile)
                {
                    debug(logger) writeln("Current line: ", line);
                    currConfig = split(line, "=");
                    debug(logger) writeln("Key: ", currConfig[0], "\nValue: ",
                            currConfig[1]);

                    switch (currConfig[0])
                    {
                        case "level" : logLevel = to!(ubyte)(currConfig[1]);
                                       break;
                        case "tzoffset" : tzoffset = to!(int)(currConfig[1]) *
                                                     ticksPerHour;
                                          break;
                        version(none)
                        {
                        case "output" : outStream = new BufferedFile(currConfig[1],
                                FileMode.OutNew);
                                        break;
                        }
                        else
                        {
                        case "output" : break;
                        }
                        default: throw new Exception("Invalid config file.");
                    }
                }

                configPoller = new Thread(&pollConfig);
                configPoller.start();
                configPoller.isDaemon(true);

                tid = cast(shared Tid)spawn(logWriter);
            }
        }
    }

    private void pollConfig()
    {
        while(1)
        {
            string[] logFile = splitlines(readText("logger.conf"));
            string[] currConfig;

            foreach (line; logFile)
            {
                currConfig = split(line, "=");
                switch (currConfig[0])
                {
                    case "level" : logLevel = to!(ubyte)(currConfig[1]);
                                   break;
                    default: break;
                }
            }

            Thread.sleep(pollTimeout);
        }
    }

    private void writeToLog(string levelName, string msg, bool doFlush = false)
    {
        send(cast(Tid)tid, levelName, msg, doFlush, this.location);
    }

    /**
        Write a string to the logfile with a level of TRACE.

        Params:
            logString = The string to be written to the logfile.

        Examples:
        ---
            lgr.trace("Value of foo: " ~ foo);
        ---
     */
    void trace(lazy string logString)
    {
        version(nologger)
        {
        }
        else
        {
            if (logLevel == 5)
            {
                writeToLog("TRACE", logString());
            }
        }
    }

    /**
        Write a string to the logfile with a level of DEBUG.

        Currently called dbug as the keyword debug is already in use.

        Params:
            logString = The string to be written to the logfile.

        Examples:
        ---
            lgr.dbug("Value of foo: " ~ foo);
        ---
     */
    void dbug(lazy string logString)
    {
        version(nologger)
        {
        }
        else
        {
            if (logLevel > 3)
            {
                writeToLog("DEBUG", logString());
            }
        }
    }

    /**
        Write a string to the logfile with a level of INFO.

        Params:
            logString = The string to be written to the logfile.

        Examples:
        ---
            lgr.info("Value of foo: " ~ foo);
        ---
     */
    void info(lazy string logString)
    {
        version(nologger)
        {
        }
        else
        {
            if (logLevel > 2)
            {
                writeToLog("INFO", logString());
            }
        }
    }

    /**
        Write a string to the logfile with a level of ERROR.

        This function forces a flush of the output buffer.

        Params:
            logString = The string to be written to the logfile.

        Examples:
        ---
            lgr.error("Value of foo: " ~ foo);
        ---
     */
    void error(lazy string logString)
    {
        version(nologger)
        {
        }
        else
        {
            if (logLevel > 1)
            {
                writeToLog("ERROR", logString(), true);
            }
        }
    }

    /**
        Write a string to the logfile with a level of FATAL.

        This function forces a flush of the output buffer.

        Params:
            logString = The string to be written to the logfile.

        Examples:
        ---
            lgr.fatal("Value of foo: " ~ foo);
        ---
     */
    void fatal(lazy string logString)
    {
        version(nologger)
        {
        }
        else
        {
            if (logLevel > 0)
            {
                writeToLog("FATAL", logString(), true);
            }
        }
    }
}

/**
    A simple logging function that, while usable, is meant to be an example
    for creating a log writer function.  The key element is the receive block.

    The current format for entries within the logfile is as follows:
    ---
    YYYY-MM-DD HH:MM:SS=LEVEL: location - msg
    ---

    This format was chosen as the time representation is simple to understand,
    the level at which the message was captured at is easy to identify, and the
    user knows where the message came from within the code.

    Examples:
    ---
        2010-02-01 14:53:27=ERROR: counter.d - Negative counter value found.
    ---
 */
// TODO: Try to make more efficient by buffering or using mmap.
void simpleLogger(string filename)
{
    Date currTime;
    bool cont = true;

    try
    {
        while (cont)
        {
            receive( (string levelName, string msg, bool doFlush,
                      string location)
            {
                currTime.parse(toUTCString(getUTCtime() + tzoffset));
                string locMsg =
                    format("%4d-%02d-%02d %02d:%02d:%02d=%s: %s - %s\n",
                    currTime.year, currTime.month, currTime.day,
                    currTime.hour, currTime.minute, currTime.second,
                    levelName, location, msg.dup);
                std.file.append(filename, locMsg);
            },
            (OwnerTerminated)
            {
                writeln("Terminating the logger.");
                cont = false;
            });
        }
    }
    catch(Exception e)
    {
        writeln("In write function: ", e.msg);
    }
}
/* alias curry!(simpleLogger, "simple.log") simpleWrite; */
